/* udp_garbage.c
 * Dissector to group together packet of udp flow that haven't a specific dissector
 *
 * $Id: udp_garbage.c,v 1.2 2007/09/08 07:13:16 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <pcap.h>

#include "proto.h"
#include "dmemory.h"
#include "etypes.h"
#include "flow.h"
#include "log.h"
#include "dnsdb.h"
#include "udp_garbage.h"
#include "pei.h"

#define GRB_FILE           0  /* to put (or not) data in to a file */
#define UDP_GRB_TMP_DIR    "udp_grb"

static int ppp_id;
static int eth_id;
static int ip_id;
static int ipv6_id;
static int udp_id;
static int ip_src_id;
static int ip_dst_id;
static int ipv6_src_id;
static int ipv6_dst_id;
static int port_src_id;
static int port_dst_id;
static int udp_grb_id;
static volatile int serial = 0;

/* pei id */
static int pei_connect_id;
static int pei_txt_id;

static volatile unsigned int incr;

static bool UdpGrbCheck(int flow_id)
{
    if (FlowPktNum(flow_id) >  UDP_GRB_PKT_LIMIT || FlowIsClose(flow_id) == TRUE) {
        return TRUE;
    }

    return FALSE;
}


static bool UdpGrbMajorityText(unsigned char *dat, unsigned int size)
{
    unsigned int perc, i, j;

    if (size == 0)
        return FALSE;

    perc = (size * UDP_GRB_PERCENTAGE)/100;
    
    j = 0;
    for (i=0; i!=size && j!=perc; i++) {
        if (0x1F<dat[i] && dat[i]<0x7F)
            j++;
    }
    if (j == perc)
        return TRUE;
    
    return FALSE;
}


static void UdpGrbText(FILE *fp, unsigned char *dat, unsigned int size)
{
    unsigned int i, j;
    
    j = 0;
    for (i=0; i!=size; i++) {
        if (dat[i]<0x7F)
            dat[j++] = dat[i];
    }
    fwrite(dat, 1, j, fp);
}


static void GrbPei(pei *ppei, const pstack_f *udp, char *txt_file, time_t *cap_sec, time_t *end_cap)
{
    char host_t[UDP_GRB_FILENAME_PATH_SIZE];
    char host[UDP_GRB_FILENAME_PATH_SIZE];
    const pstack_f *ip;
    ftval ip_host, port;
    unsigned short port_host;
    bool ipv6;
    pei_component *cmpn;
    
    ip = ProtGetNxtFrame(udp); /* ip/ipv6 frame */
    /* source */
    ProtGetAttr(udp, port_src_id, &port);
    port_host = port.uint16;
    ipv6 = FALSE;
    if (ProtFrameProtocol(ip) == ipv6_id) {
        ipv6 = TRUE;
    }
    if (ipv6 == FALSE) {
        ProtGetAttr(ip, ip_src_id, &ip_host);
        if (DnsDbSearch(&(ip_host), FT_IPv4, host_t, UDP_GRB_FILENAME_PATH_SIZE) != 0) {
            FTString(&(ip_host), FT_IPv4, host_t);
        }
    }
    else {
        ProtGetAttr(ip, ipv6_src_id, &ip_host);
        if (DnsDbSearch(&(ip_host), FT_IPv6, host_t, UDP_GRB_FILENAME_PATH_SIZE) != 0) {
            FTString(&(ip_host), FT_IPv6, host_t);
        }
    }
    sprintf(host, "udp:%s:%i", host_t, port_host);
    
    
    /* destination */
    ProtGetAttr(udp, port_dst_id, &port);
    port_host = port.uint16;
    if (ProtFrameProtocol(ip) == ipv6_id) {
        ipv6 = TRUE;
    }
    if (ipv6 == FALSE) {
        ProtGetAttr(ip, ip_dst_id, &ip_host);
        if (DnsDbSearch(&(ip_host), FT_IPv4, host_t, UDP_GRB_FILENAME_PATH_SIZE) != 0) {
            FTString(&(ip_host), FT_IPv4, host_t);
        }
    }
    else {
        ProtGetAttr(ip, ipv6_dst_id, &ip_host);
        if (DnsDbSearch(&(ip_host), FT_IPv6, host_t, UDP_GRB_FILENAME_PATH_SIZE) != 0) {
            FTString(&(ip_host), FT_IPv6, host_t);
        }
    }
    sprintf(host+strlen(host), "-%s:%i", host_t, port_host);

    /* pei components */
    PeiNewComponent(&cmpn, pei_connect_id);
    PeiCompCapTime(cmpn, *cap_sec);
    PeiCompAddStingBuff(cmpn, host);
    PeiAddComponent(ppei, cmpn);
    
    PeiNewComponent(&cmpn, pei_txt_id);
    PeiCompCapTime(cmpn, *cap_sec);
    PeiCompAddFile(cmpn, "Text", txt_file, 0);
    PeiAddComponent(ppei, cmpn);
}


packet* UdpGrbDissector(int flow_id)
{
    packet* pkt;
    const pstack_f *udp, *ip;
    ftval port_src, port_dst, ip_src, ip_dst;
    struct in_addr ip_addr;
    char ips_str[30], ipd_str[30];
    bool ipv4;
    int count;
    int threshold;
    bool txt_data;
    FILE *txt_fp;
    char txt_file[UDP_GRB_FILENAME_PATH_SIZE];
    unsigned char *thrs;
    pei *ppei;
    time_t cap_sec, end_cap;
#if GRB_FILE
    int fd_pcap;
    char filename[256];
    int prot;
    struct pcap_file_header fh;
    struct pcap_pkthdr pckt_header;
#endif

    LogPrintf(LV_DEBUG, "UDP garbage id: %d", flow_id);

    udp = FlowStack(flow_id);
    ip = ProtGetNxtFrame(udp);
    ProtGetAttr(udp, port_src_id, &port_src);
    ProtGetAttr(udp, port_dst_id, &port_dst);
    ipv4 = FALSE;
    if (ProtFrameProtocol(ip) == ip_id)
        ipv4 = TRUE;
    if (ipv4) {
        ProtGetAttr(ip, ip_src_id, &ip_src);
        ProtGetAttr(ip, ip_dst_id, &ip_dst);
        ip_addr.s_addr = ip_src.uint32;
        LogPrintf(LV_DEBUG, "\tSRC: %s:%d", inet_ntoa(ip_addr), port_src.uint16);
        sprintf(ips_str, "%s.%d", inet_ntoa(ip_addr), port_src.uint16);
        ip_addr.s_addr = ip_dst.uint32;
        LogPrintf(LV_DEBUG, "\tDST: %s:%d", inet_ntoa(ip_addr), port_dst.uint16);
        sprintf(ipd_str, "%s.%d", inet_ntoa(ip_addr), port_dst.uint16);
    }
    else
        LogPrintf(LV_DEBUG, "UDP garbage IPv6");

    /* file pcap */
#if GRB_FILE
    sprintf(filename, "%s/udp_%d_grb_%s_%s.pcap", ProtTmpDir(), serial, ips_str, ipd_str);
    serial++;
    fd_pcap = open(filename, O_WRONLY | O_CREAT, 0x01B6);
    memset(&fh, 0, sizeof(struct pcap_file_header));
    fh.magic = 0xA1B2C3D4;
    fh.version_major = PCAP_VERSION_MAJOR;
    fh.version_minor = PCAP_VERSION_MINOR;
    fh.snaplen = 65535;
    if (ProtGetNxtFrame(ip) != NULL) {
        prot = ProtFrameProtocol(ProtGetNxtFrame(ip));
        if (prot== eth_id)
            fh.linktype = DLT_EN10MB;
        else if (prot == ppp_id)
            fh.linktype = DLT_PPP;
        else
            fh.linktype = DLT_RAW;
    }
    if (fd_pcap != -1)
        write(fd_pcap, (char *)&fh, sizeof(struct pcap_file_header));
#endif

    count = 0;
    ppei = NULL;
    txt_data = FALSE;
    txt_fp = NULL;
    threshold = 0;
    thrs = xmalloc(UDP_GRB_THRESHOLD);
    pkt = FlowGetPkt(flow_id);
    if (pkt != NULL) {
        /* create pei */
        PeiNew(&ppei, udp_grb_id);
        PeiCapTime(ppei, pkt->cap_sec);
        PeiMarker(ppei, pkt->serial);
        PeiStackFlow(ppei, udp);
        cap_sec = pkt->cap_sec;
    }
    while (pkt != NULL) {
        count++;
#if GRB_FILE
        pckt_header.caplen = pkt->raw_len;
        pckt_header.len = pkt->raw_len;
        pckt_header.ts.tv_sec = pkt->cap_sec;
        pckt_header.ts.tv_usec = pkt->cap_usec;
        if (fd_pcap != -1) {
            write(fd_pcap, (char *)&pckt_header, sizeof(struct pcap_pkthdr));
            write(fd_pcap, (char *)pkt->raw, pkt->raw_len);
        }
#endif
        if (thrs != NULL) {
            /* check stream to find text */
            if (threshold + pkt->len > UDP_GRB_THRESHOLD) {
                if (txt_data == FALSE) {
                    txt_data = UdpGrbMajorityText(thrs, threshold);
                    if (txt_data == FALSE) {
                        xfree(thrs);
                        thrs = NULL;
                        threshold = 0;
                    }
                    else {
                        sprintf(txt_file, "%s/%s/udp_grb_%lu_%p_%i.txt", ProtTmpDir(), UDP_GRB_TMP_DIR, time(NULL), txt_file, incr);
                        txt_fp = fopen(txt_file, "w");
                        if (txt_fp != NULL) {
                            UdpGrbText(txt_fp, thrs, threshold);
                            threshold = 0;
                            memcpy(thrs+threshold, pkt->data,  pkt->len);
                            threshold += pkt->len;
                        }
                        else {
                            LogPrintf(LV_ERROR, "Unable to open file: %s", txt_file);
                            txt_data = FALSE;
                            xfree(thrs);
                            thrs = NULL;
                            threshold = 0;
                        }
                    }
                }
                else {
                    UdpGrbText(txt_fp, thrs, threshold);
                    threshold = 0;
                    memcpy(thrs+threshold, pkt->data,  pkt->len);
                    threshold += pkt->len;
                }
            }
            else {
                memcpy(thrs+threshold, pkt->data,  pkt->len);
                threshold += pkt->len;
            }
        }
        PktFree(pkt);
        pkt = FlowGetPkt(flow_id);
    }
    if (thrs != NULL) {
        if (txt_data == FALSE) {
            if (UdpGrbMajorityText(thrs, threshold) == TRUE) {
                sprintf(txt_file, "%s/%s/udp_grb_%lu_%p_%i.txt", ProtTmpDir(), UDP_GRB_TMP_DIR, time(NULL), txt_file, incr);
                txt_fp = fopen(txt_file, "w");
            }
        }
        if (txt_fp != NULL) {
            UdpGrbText(txt_fp, thrs, threshold);
        }
        xfree(thrs);
    }
    if (txt_fp != NULL) {
        fclose(txt_fp);
        /* insert data */
        GrbPei(ppei, udp, txt_file, &cap_sec, &end_cap);
        /* insert pei */
        PeiIns(ppei);
    }

    /* end */
#if GRB_FILE
    if (fd_pcap != -1)
        close(fd_pcap);
#endif

    LogPrintf(LV_DEBUG, "UDP garbage... bye bye  fid:%d", flow_id);

    return NULL;
}


int DissecRegist(void)
{
    proto_heury_dep hdep;
    pei_cmpt peic;

    memset(&hdep, 0, sizeof(proto_heury_dep));
    memset(&peic, 0, sizeof(pei_cmpt));

    /* protocol name */
    ProtName("UDP garbage", "udp-grb");

    /* dep: ethernet */
    hdep.name = "udp";
    hdep.ProtCheck = UdpGrbCheck;
    ProtHeuDep(&hdep);

    /* PEI components */
    peic.abbrev = "conn";
    peic.desc = "Connection";
    ProtPeiComponent(&peic);

    peic.abbrev = "txt";
    peic.desc = "Tex file";
    ProtPeiComponent(&peic);

    /* dissectors subdissectors registration */
    ProtDissectors(NULL, UdpGrbDissector, NULL, NULL);

    return 0;
}


int DissectInit(void)
{
    char tmp_dir[256];

    /* part of file name */
    incr = 0;

    /* info id */
    ppp_id = ProtId("ppp");
    eth_id = ProtId("eth");
    ip_id = ProtId("ip");
    ipv6_id = ProtId("ipv6");
    udp_id = ProtId("udp");
    ip_dst_id = ProtAttrId(ip_id, "ip.dst");
    ip_src_id = ProtAttrId(ip_id, "ip.src");
    ipv6_dst_id = ProtAttrId(ipv6_id, "ipv6.dst");
    ipv6_src_id = ProtAttrId(ipv6_id, "ipv6.src");
    port_dst_id = ProtAttrId(udp_id, "udp.dstport");
    port_src_id = ProtAttrId(udp_id, "udp.srcport");
    udp_grb_id = ProtId("udp-grb");
    
    /* pei id */
    pei_connect_id = ProtPeiComptId(udp_grb_id, "conn");
    pei_txt_id = ProtPeiComptId(udp_grb_id, "txt");
    
    /* tmp directory */
    sprintf(tmp_dir, "%s/%s", ProtTmpDir(), UDP_GRB_TMP_DIR);
    mkdir(tmp_dir, 0x01FF);
    
    return 0;
}
